/*
 * Copyright (C) 2017-2019 VIA Corporation
 * Copyright (C) 2019 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#pragma once

#include "fu-plugin.h"

#define VLI_USBHUB_PD_FLASHMAP_ADDR_LEGACY		0x4000
#define VLI_USBHUB_PD_FLASHMAP_ADDR			0x1003
