/*
 * Copyright (C) 2020 Mario Limonciello <mario.limonciello@dell.com>
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */
#pragma once

#include "fu-engine.h"

gboolean	fu_engine_update_motd		(FuEngine	*self,
						 GError		**error);
